@@ -1,26 +1,26 @@
 package Errors::Errors;
 
 ##############################################
-# Errors module written by Julian Lishev
-# This module is part of WebTools package!
-# Privacy and terms are same as WebTools!
-# See also: http://www.proscriptum.com/
+# Errors module
 ##############################################
+# Last modified: 22.01.2003
 
 use strict;
 
-# ----- Global members for all objects -----
+# ----- Global members (public) for all objects -----
 $Errors::Errors::debugging = 0;
 $Errors::Errors::sys_last_TERM_obj = '';
 $Errors::Errors::sys_last_ALRM_obj = '';
 $Errors::Errors::sys_sent_content  = 0;
 $Errors::Errors::sys_last_ERROR = 0;
 $Errors::Errors::sys_exit_called = 0;
+$Errors::Errors::sys_last_term_invoked = 0;
+$Errors::Errors::sys_last_alrm_invoked = 0;
 
 BEGIN
  {
   use vars qw($VERSION @ISA @EXPORT);
-  $VERSION = "1.01";
+  $VERSION = "1.02";
   @ISA = qw(Exporter);
   @EXPORT = qw();
  }
@@ -107,30 +107,44 @@ sub __default_header
 
 sub __default_SIGNALS
 {
+  if($Errors::Errors::sys_last_term_invoked == 1) {return(0);}  # Function already invoked! Can't proceed that request!
+  $Errors::Errors::sys_last_term_invoked = 1;
+  my $how = shift;
   my $self = $Errors::Errors::sys_last_TERM_obj;
   my $sub;
   my $hashref = $self->{'__errors'};
   foreach $sub (keys %$hashref)
    {
     my $ref = $hashref->{$sub}->{'onterm'};
-    if(ref($ref)) { &$ref($self,'term',''); }
+    if(ref($ref)) { &$ref($self,('type'=>'term','name'=>$sub,'signal'=>$how)); }
    }
   my $ref = $self->{'__subs'}->{'onterm'};
-  if(ref($ref)) { &$ref($self,'term',''); }
+  if(ref($ref)) { &$ref($self,('type'=>'term','name'=>'','signal'=>$how)); }
+  $Errors::Errors::sys_last_term_invoked = 0;     # Done..it's now safe to make errors ;-)
 }
 
+sub __default_SIGNALS_TERM { &Errors::Errors::__default_SIGNALS('term'); }
+sub __default_SIGNALS_QUIT { &Errors::Errors::__default_SIGNALS('quit'); }
+sub __default_SIGNALS_PIPE { &Errors::Errors::__default_SIGNALS('pipe'); }
+sub __default_SIGNALS_STOP { &Errors::Errors::__default_SIGNALS('stop'); }
+sub __default_SIGNALS_HUP  { &Errors::Errors::__default_SIGNALS('hup'); }
+sub __default_SIGNALS_INT  { &Errors::Errors::__default_SIGNALS('int'); }
+
 sub __default_ALRM
 {
+  if($Errors::Errors::sys_last_alrm_invoked == 1) {return(0);}
+  $Errors::Errors::sys_last_alrm_invoked = 1;
   my $self = $Errors::Errors::sys_last_ALRM_obj;
   my $sub;
   my $hashref = $self->{'__errors'};
   foreach $sub (keys %$hashref)
    {
     my $ref = $hashref->{$sub}->{'ontimeout'};
-    if(ref($ref)) { &$ref($self,'timeout',''); }
+    if(ref($ref)) { &$ref($self,('type'=>'term','name'=>$sub,'signal'=>'alrm')); }
    }
   my $ref = $self->{'__subs'}->{'ontimeout'};
-  if(ref($ref)) { &$ref($self,'timeout',''); }
+  if(ref($ref)) { &$ref($self,('type'=>'term','name'=>'','signal'=>'alrm')); }
+  $Errors::Errors::sys_last_alrm_invoked = 0;
 }
 
 sub install
@@ -141,15 +155,17 @@ sub install
  my $sub  = shift;
  $name = lc($name);
  $sub  = lc($sub);
- 
+
  if($sub eq '')
   {
    if($name eq 'onterm')
     {
-     $SIG{'TERM'} = \&__default_SIGNALS;
-     $SIG{'QUIT'} = \&__default_SIGNALS;
-     $SIG{'PIPE'} = \&__default_SIGNALS;
-     $SIG{'STOP'} = \&__default_SIGNALS;
+     $SIG{'TERM'} = \&__default_SIGNALS_TERM;
+     $SIG{'QUIT'} = \&__default_SIGNALS_QUIT;
+     $SIG{'PIPE'} = \&__default_SIGNALS_PIPE;
+     $SIG{'STOP'} = \&__default_SIGNALS_STOP;
+     $SIG{'HUP'}  = \&__default_SIGNALS_HUP;
+     $SIG{'INT'}  = \&__default_SIGNALS_INT;
      $Errors::Errors::sys_last_TERM_obj = $self;
     }
    if($name eq 'ontimeout')
@@ -178,10 +194,12 @@ sub uninstall
   {
    if($name eq 'onterm')
     {
-     $SIG{'TERM'} = \&__default_SIGNALS;
-     $SIG{'QUIT'} = \&__default_SIGNALS;
-     $SIG{'PIPE'} = \&__default_SIGNALS;
-     $SIG{'STOP'} = \&__default_SIGNALS;
+     $SIG{'TERM'} = \&__default_SIGNALS_TERM;
+     $SIG{'QUIT'} = \&__default_SIGNALS_QUIT;
+     $SIG{'PIPE'} = \&__default_SIGNALS_PIPE;
+     $SIG{'STOP'} = \&__default_SIGNALS_STOP;
+     $SIG{'HUP'}  = \&__default_SIGNALS_HUP;
+     $SIG{'INT'}  = \&__default_SIGNALS_INT;
     }
    if($name eq 'ontimeout')
     {
@@ -273,6 +291,7 @@ sub error
   my $self  = shift;
   my $value = shift;
   my $name  = shift;
+  my @d = @_;
   my @res = ();
   my $sub;
   if($name eq '')
@@ -298,10 +317,10 @@ sub error
   foreach $sub (keys %$hashref)
    {
     my $ref = $hashref->{$sub}->{'onerror'};
-    if(ref($ref)) { &$ref($self,$value,$name,@_); }
+    if(ref($ref)) { &$ref($self,('type'=>'error','error'=>$value,'name'=>$sub,'to'=>$name,'params'=>\@d)); }
    }
   my $ref = $self->{'__subs'}->{'onerror'};
-  if(ref($ref)) { &$ref($self,$value,$name,@_); }
+  if(ref($ref)) { &$ref($self,('type'=>'error','error'=>$value,'name'=>'','to'=>$name,'params'=>\@d)); }
  
   return(@res);
 }
@@ -309,6 +328,7 @@ sub error
 sub die
 {
   my $self = shift;
+  my @d = @_;
   if($self->{'__counters'}->{'die'} == 0)
   {
    $Errors::Errors::sys_exit_called = 1;
@@ -318,18 +338,19 @@ sub die
    foreach $sub (keys %$hashref)
     {
      my $ref = $hashref->{$sub}->{'onexit'};
-     if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','die'); }
+     if(ref($ref)) { &$ref($self,('type'=>'die','error'=>$Errors::Errors::sys_last_ERROR,'name'=>$sub,'params'=>\@d)); }
     }
    my $ref = $self->{'__subs'}->{'onexit'};
-   if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','die'); }
+   if(ref($ref)) { &$ref($self,('type'=>'die','error'=>$Errors::Errors::sys_last_ERROR,'name'=>'','params'=>\@d)); }
    die(@_);
   }
-  else {die(@_);}
+  else {return(0);}
 }
 
 sub exit
 {
   my $self = shift;
+  my @d = @_;
   if($self->{'__counters'}->{'exit'} == 0)
   {
    $Errors::Errors::sys_exit_called = 1;
@@ -339,18 +360,19 @@ sub exit
    foreach $sub (keys %$hashref)
     {
      my $ref = $hashref->{$sub}->{'onexit'};
-     if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','exit'); }
+     if(ref($ref)) { &$ref($self,('type'=>'exit','error'=>$Errors::Errors::sys_last_ERROR,'name'=>$sub,'params'=>\@d)); }
     }
    my $ref = $self->{'__subs'}->{'onexit'};
-   if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','exit'); }
+   if(ref($ref)) { &$ref($self,('type'=>'exit','error'=>$Errors::Errors::sys_last_ERROR,'name'=>'','params'=>\@d)); }
    exit(@_);
   }
-  else {exit(@_);}
+  else {return(0);}
 }
 
 sub DESTROY
 {
   my $self = shift;
+  my @d = ();
   my $sub;
   my $hashref = $self->{'__errors'};
   if(!$Errors::Errors::sys_exit_called)
@@ -361,10 +383,10 @@ sub DESTROY
     foreach $sub (keys %$hashref)
      {
       my $ref = $hashref->{$sub}->{'onexit'};
-      if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','destroy'); }
+      if(ref($ref)) { &$ref($self,('type'=>'destroy','error'=>$Errors::Errors::sys_last_ERROR,'name'=>$sub,'params'=>\@d)); }
      }
     my $ref = $self->{'__subs'}->{'onexit'};
-    if(ref($ref)) { &$ref($self,$Errors::Errors::sys_last_ERROR,'','destroy'); }
+    if(ref($ref)) { &$ref($self,('type'=>'destroy','error'=>$Errors::Errors::sys_last_ERROR,'name'=>'','params'=>\@d)); }
    }
 }
 
@@ -386,80 +408,108 @@ Error module is created as base "error" catcher module especially for Web
 =head1 SYNOPSIS
 
  use Errors::Errors;
+ 
+ use strict;
 
- $obj = Errors::Errors->new();
+ my $obj = Errors::Errors->new();
  
  $obj->content(0);
  $obj->header();
 
- $obj->attach('xreader');  # Attach sub object for error of type 'xreader'
- $obj->attach('myown');    # Attach sub object for error of type 'myown'
+ $obj->attach('xreader');  # Attach sub object (sobject) for error of type 'xreader'.
+ $obj->attach('myown');    # Attach sub object for error of type 'myown'.
  
- $hash = {
-	  name=>'July',
-	  born_year=>'81',
-	 };
+ my $hash = {
+	     name=>'July',
+	     born_year=>'81',
+	    };
  
  $obj->attach_object('xreader',$hash); # Hash ref or object
  
- $obj->install('onTerm',\&custom);
- $obj->install('onError',\&anysub,'xreader');
- $obj->install('onExit',\&leave);
- $obj->install('onTerm',\&custom,'myown');
+ $obj->install('onTerm',\&custom);            # Install sub for term event.
+ $obj->install('onError',\&anysub,'xreader'); # Install sub for xreader sobject which will be called when error occure.
+ $obj->install('onExit',\&leave);             # Install sub for exit event.
+ $obj->install('onTerm',\&custom,'myown');    # Install additional term sub for 'myown' sobject.
  
- $obj->error(7,'xreader');
+ $obj->error(7,'xreader');  # Set error '7' for 'xreader' sobject and force execution of error chain.
  
- $h = $obj->fetch_object('xreader');
- $obj->print($h->{name});
+ #my $h = $obj->fetch_object('xreader');  # Fetch attached object from xreader.
+ #$obj->print($h->{name}."\n");
  
- $obj->uninstall('onError','xreader');
+ $obj->uninstall('onError','xreader');  # Remove 'xreader' sub from 'error' chain.
  
- $obj->detach_object('xreader');
- $obj->detach('xreader');
+ $obj->detach_object('xreader'); # Remove attached object from xreader.
+ $obj->detach('xreader'); # Remove xreader (and attached objects).
  
- $obj->install('onTimeOut',\&timeout);
- eval 'alarm(2);';
+ $obj->install('onTimeOut',\&timeout); # Install timeout sub routine.
+ eval 'alarm(1);';    # Force alarm after 1s
+ sleep(2);            # Wait 2 sec, so alarm will be activeted and 'timeout' chain will be turned.
+ #Note: Press CTRL+C to active 'INT' signal ('onTerm' event) before time to flow out.
  
- $obj->exit();
+ #$obj->exit();       # Force exit of program.
+ #$obj->die();        # Force die of program.
+ 
+ #By default script close 'normally', so 'destroy' chain will be executed.
 
  sub custom {
-  my $obj   = shift;       # 'Errors' object
-  my $err   = shift;       # Error number/message (for TERM it has value 'term')
-  my $name  = shift;       # 'name' of error (for TERM it has empty value)
+  my $obj   = shift;         # 'Errors' object
+  my %in    = @_;
+  my $type  = $in{'type'};   # Error type: 'term'
+  my $name  = $in{'name'};   # Custom error name
+  my $sig   = $in{'signal'}; # Invoked signal (term,quit,pipe...)
   # ...blah...blah...
+  print "Custom($name)\n";
  }
  sub leave {
-  my $obj   = shift;       # 'Errors' object
-  my $err   = shift;       # Last error number/message
-  my $name  = shift;       # 'name' of error
-  my $how   = shift;       # can be: 'exit','die' or 'destroy'
+  my $obj    = shift;
+  my %in     = @_;
+  my $type   = $in{'type'};   # Error type: 'exit','die','destroy'
+  my $err    = $in{'error'};  # Error value
+  my $name   = $in{'name'};   # Custom error name
+  my $params = $in{'params'}; # Additional parameters (ref to @)
+  my @params = @$params;
   # ...blah...blah...
+  print "$type\n";
  }
  sub timeout
  {
-  my $obj    = shift;      # 'Errors' object
-  my $what   = shift;      # 'timeout' string
+  my $obj   = shift;
+  my %in    = @_;
+  my $type  = $in{'type'};    # Error type: 'term'
+  my $name  = $in{'name'};    # Custom error name
+  my $sig   = $in{'signal'};  # Invoked signal: 'alrm'
   # ...blah...blah...
-  print "Time OUT";
+  print "Timeout\n";
  }
  sub anysub {
-  my $obj   = shift;       # 'Errors' object
-  my $err   = shift;       # Error number/message
-  my $name  = shift;       # 'name' of error
-  if($name =~ m/xreader/si)
+  my $obj    = shift;
+  my %in     = @_;
+  my $type   = $in{'type'};   # Error type: 'error'
+  my $err    = $in{'error'};  # Error message
+  my $name   = $in{'name'};   # Custom error name
+  my $to     = $in{'to'};     # Message is sent "to"
+  my $params = $in{'params'}; # Additional parameters (ref to @)
+  my @params = @$params;
+  
+  
+  if($name eq $to && $to eq 'xreader')
    {
-    $obj->print ("Error in Xreader!!!");  # If error is raised in 'xreader'
+    $obj->print ("Error in Xreader!!!\n");  # If error is raised in 'xreader'
+    my $h = $obj->fetch_object('xreader');
+    $obj->print ($h->{born_year}."\n");
    }
   else
    {
-    $obj->print ("Error in ... I don't know :-)!!!");
+    $obj->print ("Error in ... I don't know ;-)!!!\n");
    }
  }
 
 =head1 AUTHOR
 
- Julian Lishev - Bulgaria, Sofia, 
- e-mail: julian@proscriptum.com, 
- www.proscriptum.com
+ Julian Lishev - Bulgaria,Sofia
+ e-mail: julian@proscriptum.com
+
+ Copyright (c) 2001, Julian Lishev, Sofia 2003
+ All rights reserved.
 
 =cut
